#include "../include/COLREGs.hpp"
#include "../include/misc.hpp"

COLREGs::COLREGs(std::vector<std::pair<PoseSE2,Eigen::Vector3d>>& ship_motion_data, int& ownship_index)
{
    t_low = 1;
    t_high = 20;
    d = 20;

    t_hysteresis = 0.1;
    d_hysteresis = 0.1;

    epsilon = 0.01;

    d_range = 200;
    d_close = 14;

    motion_data = ship_motion_data;
    eta = motion_data[ownship_index].first;
    nu = motion_data[ownship_index].second;

    motion_data.erase(motion_data.begin() + ownship_index);

    p = eta.position();
    v = (R_psi(eta.theta()) * nu).head(2);
}


void COLREGs::updateState(bool& safe)
{
    for (int i = 0; i < motion_data.size(); ++i)
    {
        PoseSE2 eta_o = motion_data[i].first;
        Eigen::Vector3d nu_o = motion_data[i].second;
        Eigen::Vector2d p_o = eta_o.position();
        Eigen::Vector2d v_o = (R_psi(eta_o.theta()) * nu_o).head(2);

        double TCPA = ((v - v_o).norm() <= epsilon)? 0 : (p_o - p).dot(v - v_o) / (v - v_o).squaredNorm();
        double DCPA = ((p + TCPA * v) - (p_o + TCPA * v_o)).norm();

        if (safe)
        {   
            if ( (DCPA <= d  && inInterval(TCPA, t_low , t_high)) )
            {
                safe = false;
                return;
            }
        }
        else
        {
            if ( DCPA <= d + d_hysteresis && inInterval(TCPA, t_low - t_hysteresis, t_high + t_hysteresis) )
            {
                safe = false;
                return;
            }
        }

        if ((p - p_o).norm() <= d_close)
        {
            safe = false;
            return;
        }
    }

    safe = true;
}


int COLREGs::findSector(double& angle, double rotation)
{
    int flag;
    double theta_1 = M_PI/8;
    double theta_2 = 5*M_PI/8;

    auto intervals = std::vector<std::pair<double,double>>
    {
        {g2o::normalize_theta(rotation-theta_1), g2o::normalize_theta(rotation+theta_1)},
        {g2o::normalize_theta(rotation+theta_1), g2o::normalize_theta(rotation+theta_2)},
        {g2o::normalize_theta(rotation+theta_2), g2o::normalize_theta(rotation-theta_2)},
        {g2o::normalize_theta(rotation-theta_2), g2o::normalize_theta(rotation-theta_1)}
    };

    for (int i=0; i<intervals.size(); ++i)
    {
        if (intervals[i].first > intervals[i].second)
            {
                flag = i;
                break;
            }
    }

    for (int i=0; i<intervals.size(); ++i)
    {
        if ( flag == i && ( inInterval(angle, intervals[i].first, M_PI) || inInterval(angle, -M_PI, intervals[i].second)) )
            return i;
        else if ( inInterval(angle, intervals[i].first, intervals[i].second) )
            return i;
    }

    return -1;
}


void COLREGs::updateEncounterType()
{
    encounter_types.clear();

    auto R1 = std::vector<std::string>{"HO","GW","OT","SO"};
    auto R2 = std::vector<std::string>{"GW","GW","SF","SF"};
    auto R3 = std::vector<std::string>{"SF","SF","SF","SF"};
    auto R4 = std::vector<std::string>{"SO","SF","SF","SO"};

    auto COLREGs = std::vector<std::vector<std::string>>{R1, R2, R3, R4};

    for (int i = 0; i < motion_data.size(); ++i)
    {
        PoseSE2 eta_o = motion_data[i].first;
        Eigen::Vector2d p_o = eta_o.position();
        Eigen::Vector2d p_r = p_o - p;

        double bearing = atan2(p_r(1), p_r(0));
        double bearing_r = g2o::normalize_theta(bearing - eta.theta());
        double heading_r = g2o::normalize_theta(eta_o.theta() - eta.theta());

        if ( (p - p_o).norm() < d_range )
        {
            int region = findSector(bearing_r, 0);
            int sector = findSector(heading_r, bearing_r - M_PI);          
            auto encounter_type = COLREGs[region][sector];
            if ( (encounter_type == "SF"||encounter_type == "SO") && (p - p_o).norm() < d_close )
                encounter_type = "CS";
            encounter_types.push_back(encounter_type);
        }
    }
}



std::pair<bool,RotType> COLREGs::getAction()
{
    RotType turning_dir = RotType::none;
    bool SO = false;

    for (int i=0; i < encounter_types.size(); ++i)
    {
        if (encounter_types[i] == "SO")
        {
            SO = true;
            break;
        }   
    }

    for (int i=0; i < encounter_types.size(); ++i)
    {
        if (encounter_types[i] == "HO" || encounter_types[i] == "GW")
        {
            turning_dir = RotType::right;
            SO = false;
            break;
        }

        if (encounter_types[i] == "OT" || encounter_types[i] == "CS")
            SO = false;
    }

    return {SO, turning_dir};
}