#include "../include/Guidance.hpp"
#include <g2o/stuff/misc.h>
#include <Eigen/Dense>

Guidance::Guidance(std::vector<Eigen::Vector2d>& waypoints)
{
    sampleTime = 0.02;
    idx = 1;
    goal_dist = 30;
    switch_dist = 20;
    y_e_int = 0;
    Kp = 0.02;
    Ki = 0;

    ref.r_max = 20 * M_PI / 180;
    ref.alpha_max = ref.r_max / 5;
    ref.w_psi_d = 1;
    ref.zeta_psi_d = 1;

    ref.u_max = 3;
    ref.a_max = ref.u_max / 10;
    ref.w_u_d = 1.5;
    ref.zeta_u_d = 1;

    cref.r_max = 10 * M_PI / 180;
    cref.alpha_max = ref.r_max / 10;
    cref.w_psi_d = 10;
    cref.zeta_psi_d = 1;

    cref.u_max = 2.5;
    cref.a_max = ref.u_max / 10;
    cref.w_u_d = 10;
    cref.zeta_u_d = 1;

    waypoints_ = waypoints;
    reached_goal = false;
}


void Guidance::headingRefModel(double& psi_ref, double& psi_d, double& r_d, double& alpha_d)
{
    r_d = g2o::sign(r_d) * std::min(fabs(r_d), ref.r_max);
    alpha_d = g2o::sign(alpha_d) * std::min(fabs(alpha_d), ref.alpha_max);
    
    double j_d = pow(ref.w_psi_d, 3) * g2o::normalize_theta(psi_ref-psi_d) - (2*ref.zeta_psi_d+1) * pow(ref.w_psi_d, 2) * r_d - (2*ref.zeta_psi_d+1) * ref.w_psi_d * alpha_d;

    psi_d += sampleTime * r_d;
    psi_d = g2o::normalize_theta(psi_d);
    r_d += sampleTime * alpha_d;
    alpha_d += sampleTime * j_d;
}


void Guidance::speedRefModel(double& u_ref, double& u_d, double& a_d)
{
    u_d = g2o::sign(u_d) * std::min(fabs(u_d), ref.u_max);
    a_d = g2o::sign(a_d) * std::min(fabs(a_d), ref.a_max);

    double j_d = pow(ref.w_u_d, 2) * (u_ref-u_d) - 2 * ref.w_u_d * ref.zeta_u_d * a_d;

    u_d += sampleTime * a_d;
    a_d += sampleTime * j_d;
}

void Guidance::COLAVheadingRefModel(double& psi_ref, double& psi_d, double& r_d, double& alpha_d)
{
    r_d = g2o::sign(r_d) * std::min(fabs(r_d), cref.r_max);
    alpha_d = g2o::sign(alpha_d) * std::min(fabs(alpha_d), cref.alpha_max);
    
    double j_d = pow(cref.w_psi_d, 3) * g2o::normalize_theta(psi_ref-psi_d) - (2*cref.zeta_psi_d+1) * pow(cref.w_psi_d, 2) * r_d - (2*cref.zeta_psi_d+1) * cref.w_psi_d * alpha_d;

    psi_d += sampleTime * r_d;
    psi_d = g2o::normalize_theta(psi_d);
    r_d += sampleTime * alpha_d;
    alpha_d += sampleTime * j_d;
}


void Guidance::COLAVspeedRefModel(double& u_ref, double& u_d, double& a_d)
{
    u_d = g2o::sign(u_d) * std::min(fabs(u_d), cref.u_max);
    a_d = g2o::sign(a_d) * std::min(fabs(a_d), cref.a_max);

    double j_d = pow(cref.w_u_d, 2) * (u_ref-u_d) - 2 * cref.w_u_d * cref.zeta_u_d * a_d;

    u_d += sampleTime * a_d;
    a_d += sampleTime * j_d;
}

double Guidance::LOS(PoseSE2& eta)
{   
    if ((eta.position() - waypoints_[idx]).norm() < switch_dist)
    {
        if (idx < waypoints_.size()-1)
            idx += 1;
        else if (idx == waypoints_.size()-1)
            reached_goal = true;
    }

    double x = eta.x();
    double y = eta.y();

    double x_t = waypoints_[idx](0);
    double y_t = waypoints_[idx](1);

    double x_ref = waypoints_[idx-1](0);
    double y_ref = waypoints_[idx-1](1);

    double Pi_p = atan2(y_t - y_ref, x_t - x_ref);

    Eigen::Vector3d b_;
    Eigen::Matrix3d A_;
    Eigen::Vector3d x_;

    b_ << cos(Pi_p) * x + sin(Pi_p) * y,
         -sin(Pi_p) * x + cos(Pi_p) * y,
          tan(Pi_p) * x_t - y_t;

    A_ << cos(Pi_p), sin(Pi_p), 0,
         -sin(Pi_p), cos(Pi_p), 1,
          tan(Pi_p),        -1, 0;

    x_ = A_.inverse() * b_;

    double y_e = x_(2);

    double psi_d = g2o::normalize_theta(Pi_p - atan(Kp * y_e + Ki * y_e_int));
    y_e_int += sampleTime * y_e;

    return psi_d;
}

std::pair<PoseSE2, Eigen::Vector3d> Guidance::getLocalGoal(PoseSE2& eta, double& u_ref)
{
    if ((eta.position() - waypoints_[idx]).norm() < switch_dist)
    {
        if (idx < waypoints_.size()-1)
            idx += 1;
        else if (idx == waypoints_.size()-1)
            reached_goal = true;
    }
    
    double x = eta.x();
    double y = eta.y();

    double x_t = waypoints_[idx](0);
    double y_t = waypoints_[idx](1);

    double x_ref = waypoints_[idx-1](0);
    double y_ref = waypoints_[idx-1](1);

    double Pi_p = atan2(y_t - y_ref, x_t - x_ref);

    Eigen::Vector3d b_;
    Eigen::Matrix3d A_;
    Eigen::Vector3d x_;

    b_ << cos(Pi_p) * x + sin(Pi_p) * y,
         -sin(Pi_p) * x + cos(Pi_p) * y,
          tan(Pi_p) * x_t - y_t;

    A_ << cos(Pi_p), sin(Pi_p), 0,
         -sin(Pi_p), cos(Pi_p), 1,
          tan(Pi_p),        -1, 0;

    x_ = A_.inverse() * b_;

    double x_p = x_(0);
    double y_p = x_(1);

    Eigen::Vector2d dir_vec(x_t - x_ref, y_t - y_ref);
    Eigen::Vector2d unit_vec = dir_vec/dir_vec.norm();

    PoseSE2 pose_goal = PoseSE2(x_p, y_p, Pi_p) + goal_dist * PoseSE2(unit_vec, 0);

    Eigen::Vector3d vel_goal(u_ref, 0, 0);

    if ((eta.position() - waypoints_[idx]).norm() < goal_dist)
        pose_goal = PoseSE2(waypoints_[idx], Pi_p);

    if (reached_goal)
        vel_goal << 0, 0, 0;

    return {pose_goal, vel_goal};
}