#include "../include/Motion_Prediction.hpp"
#include "../include/misc.hpp"

MotionPrediction::MotionPrediction(std::vector<std::pair<PoseSE2,Eigen::Vector3d>>& ship_motion_data, int& ownship_index)
{
    timeStep = 2;
    d_range = 50;
    ownship_idx = ownship_index;
    motion_data = ship_motion_data;
    p = motion_data[ownship_idx].first.position();
    motion_data.erase(motion_data.begin() + ownship_idx);
    prediction_horizon = 10;
}


void MotionPrediction::straightLineMotion(std::pair<PoseSE2,Eigen::Vector3d>& ship_motion)
{
    PoseSE2 eta = ship_motion.first;
    Eigen::Vector3d nu = ship_motion.second;

    Eigen::Vector3d eta_dot = R_psi(eta.theta()) * nu;
    PoseSE2 eta_dot_(eta_dot(0),eta_dot(1),eta_dot(2));
    PoseSE2 eta_predicted = eta + eta_dot_ * timeStep;

    ship_motion = std::make_pair(eta_predicted, nu);
}


void MotionPrediction::circularMotion(std::pair<PoseSE2,Eigen::Vector3d>& ship_motion)
{
    PoseSE2 eta = ship_motion.first;
    Eigen::Vector3d nu = ship_motion.second;

    double U = sqrt(pow(nu(0),2) + pow(nu(1),2));
    double r = nu(2);
    double beta_c = atan(nu(1)/nu(0));
    double chi = eta.theta() + beta_c;
    double chi_predicted = chi + r * timeStep;

    PoseSE2 eta_predicted(eta.x() + U/r * (sin(chi_predicted) - sin(chi)),
                          eta.y() + U/r * (-cos(chi_predicted) + cos(chi)),
                          eta.theta() + r * timeStep);

    ship_motion = std::make_pair(eta_predicted, nu);
}


std::vector<Eigen::Vector2d> MotionPrediction::predict()
{
    std::vector<Eigen::Vector2d> predicted_position;

    std::vector<double>
    u_varied {0},
    r_varied {0};

    for (int i = 0; i < motion_data.size(); ++i)
    {
        auto motion = motion_data[i];

        if ( (p - motion.first.position()).norm() < d_range )
        {
            predicted_position.push_back(motion.first.position());

            for (int m = 0; m < u_varied.size(); ++m)
            {
                for (int n = 0; n < r_varied.size(); ++n)
                {
                    auto motion_varied = motion;
                    motion_varied.second[0] += u_varied[m];
                    motion_varied.second[2] += r_varied[n];

                    for (int j=0; j < int(prediction_horizon/timeStep); ++j)
                    {
                        if (motion_varied.second[0] == 0 || motion_varied.second[2] == 0)
                            straightLineMotion(motion_varied);
                        else
                            circularMotion(motion_varied);
                        
                        predicted_position.push_back(motion_varied.first.position());
                    }
                }
            }
        }
    }
    
    return predicted_position;
}

