#include "../include/Tanker.hpp"
#include <cmath>
#include <g2o/stuff/misc.h>
#include <iostream>


using namespace teb_local_planner;

Tanker::Tanker()
{   
    L = 304.8;
    T = 18.46 ;
    h = 100;
    delta_max = 30 * M_PI / 180;
    Ddelta_max = 5 * M_PI / 180;
    n_max = 90.0 / 60;
    Dn_max = 10.0 / 60;
    t = 0.22;
    cun = 0.605;
    cnn = 38.2;
    Tuu = -0.00695;
    Tun = -0.00063;
    Tnn = 0.0000354;
    d11 = 2.020;  
    d22 = -0.752; 
    d33 = -0.231;
    YT = 0.04;
    NT = -0.02;
    Xuu = -0.0377;
    Yvv = -2.400;
    Nvr = -0.300;
    Xvv = 0.3;
    Yuv = -1.205;
    Nuv = -0.451;
    Xudotz = -0.05;
    Yvdotz = -0.387;
    Nrdotz = -0.0045;
    Xuuz = -0.0061;
    Yurz = 0.182;
    Nurz = -0.047;
    Xvrz = 0.387;
    Yvvz = -1.5;
    Nvrz = -0.120;
    Xccdd = -0.093;
    Yuvz = 0.0;
    Nuvz = -0.241;
    Xccbd = 0.152;
    Yccd = 0.208;
    Nccd = -0.098;
    Xvvzz = 0.0125;
    Yccbbd = -2.16;
    Nccbbd = 0.688;
    Yccbbdz = -0.191;
    Nccbbdz = 0.344;
    z = T / (h - T);
    m11 = 1.050 - Xudotz * z;
    m22 = 2.020 - Yvdotz * z;
    m33 = 0.1232 - Nrdotz * z;

    control.m_psi = 500;
    control.d_psi = 0;
    control.k_psi = 0;
    control.e_psi_i = 0;
    control.wn_psi = 0.15;
    control.zeta_psi = 1;
    control.K_sigma = 0.001;
    control.phi = 100;
    control.e_u_i = 0;

    sampleTime = 0.1;
}


void Tanker::headingControl(double& delta_c, double psi, double psi_d, double r, boost::optional<double> r_d)
{
    double e_psi = g2o::normalize_theta(psi - psi_d);
    double e_r = (r_d) ? r - *r_d : r;
    
    double Kp = control.m_psi * pow(control.wn_psi, 2) - control.k_psi;
    double Kd = control.m_psi * 2 * control.zeta_psi * control.wn_psi - control.d_psi;
    double Ki = (control.wn_psi / 10) * Kp;

    delta_c = Kp * e_psi + Kd * e_r + Ki * control.e_psi_i;
    delta_c = g2o::sign(delta_c) * std::min(fabs(delta_c), delta_max);

    if (fabs(delta_c) < delta_max) control.e_psi_i += sampleTime * e_psi;
}


void Tanker::speedControl(double& n_c, double u, double u_d, boost::optional<double> u_d_dot)
{   
    double e_u = u - u_d;
    double sigma = e_u + control.e_u_i;

    double gT = (u_d_dot) ? 
                m11 / (1-t) * ( *u_d_dot + u_d - u - (Xuu + Xuuz * z)*pow(u, 2)/(m11*L) - control.K_sigma * tanh(sigma/control.phi) ):
                m11 / (1-t) * ( u_d - u - (Xuu + Xuuz * z)*pow(u, 2)/(m11*L) - control.K_sigma * tanh(sigma/control.phi) );
        
    double a = L * Tnn;
    double b = Tun * pow(u, 2);
    double c = Tuu * pow(u, 2) / L - gT;

    if ( pow(b, 2) - 4*a*c < 0 )
    {
        n_c = n_c;
        //std::cout <<  "Warning: radicant is smaller than 0." << std::endl;
    }
    else
    {
        n_c = (sqrt(pow(b, 2) - 4*a*c) - b) / (2*a);
        n_c = g2o::sign(n_c) * std::min(fabs(n_c), n_max);
    }

    if (fabs(n_c) < n_max) control.e_u_i += sampleTime * e_u;
}


void Tanker::dynamics(Eigen::Vector3d& nu, double& delta, double& n, double delta_c, double n_c)
{
    double u = nu.coeff(0);
    double v = nu.coeff(1);
    double r = nu.coeff(2);
    double beta = (u != 0) ? v/u : 0;

    double gT = 1 / L * Tuu * pow(u, 2) + Tun * u * n + L * Tnn * fabs(n) * n;
    double c = sqrt(cun * u * n + cnn * pow(n, 2));

    double gX = (1 / L *
                ( Xuu * pow(u, 2)
                + L * d11 * v * r
                + Xvv * pow(v, 2)
                + Xccdd * fabs(c) * c * pow(delta, 2)
                + Xccbd * fabs(c) * c * beta * delta
                + L * gT * (1 - t)
                + Xuuz * pow(u, 2) * z
                + L * Xvrz * v * r * z
                + Xvvzz * pow(v, 2) * pow(z, 2)));

    double gY = (1 / L *
                ( Yuv * u * v
                + Yvv * fabs(v) * v
                + Yccd * fabs(c) * c * delta
                + L * d22 * u * r
                + Yccbbd * fabs(c) * c * fabs(beta) * beta * fabs(delta)
                + YT * gT * L
                + L * Yurz * u * r * z
                + Yuvz * u * v * z
                + Yvvz * fabs(v) * v * z
                + Yccbbdz * fabs(c) * c * fabs(beta) * beta * fabs(delta) * z));

    double gLN =  Nuv * u * v
                + L * Nvr * fabs(v) * r
                + Nccd * fabs(c) * c * delta
                + L * d33 * u * r
                + Nccbbd * fabs(c) * c * fabs(beta) * beta * fabs(delta)
                + L * NT * gT
                + L * Nurz * u * r * z
                + Nuvz * u * v * z
                + L * Nvrz * fabs(v) * r * z
                + Nccbbdz * fabs(c) * c * fabs(beta) * beta * fabs(delta) * z;

    Eigen::Vector3d nu_dot ( gX/m11 , gY/m22 , gLN/(pow(L, 2) * m33) );

    delta = g2o::sign(delta) * std::min(fabs(delta), delta_max);
    n = g2o::sign(n) * std::min(fabs(n), n_max);

    double delta_dot = delta_c - delta;
    double n_dot = n_c - n;

    delta_dot = g2o::sign(delta_dot) * std::min(fabs(delta_dot), Ddelta_max);
    n_dot = g2o::sign(n_dot) * std::min(fabs(n_dot), Dn_max);

    nu += sampleTime * nu_dot;
    delta += sampleTime * delta_dot;
    n += sampleTime * n_dot;
}

void Tanker::kinematics(PoseSE2& eta, Eigen::Vector3d nu)
{   
    double psi = eta.theta();
    Eigen::Matrix3d R;
    R << cos(psi), -sin(psi), 0,
         sin(psi),  cos(psi), 0,
                0,         0, 1;

    Eigen::Vector3d eta_dot_ = R * nu;
    PoseSE2 eta_dot = PoseSE2(eta_dot_.coeff(0), eta_dot_.coeff(1), eta_dot_.coeff(2));

    eta += sampleTime * eta_dot;
}

void Tanker::simulate(PoseSE2& eta, Eigen::Vector3d& nu, double& delta, double& n, double& delta_c, double& n_c, double psi_d, double u_d, boost::optional<double> r_d, boost::optional<double> u_d_dot)
{
    double u = nu.coeff(0);
    double r = nu.coeff(2);
    double psi = eta.theta();

    (r_d) ? headingControl(delta_c, psi, psi_d, r, r_d) : headingControl(delta_c, psi, psi_d, r, boost::none);
    (u_d_dot) ? speedControl(n_c, u, u_d, u_d_dot) : speedControl(n_c, u, u_d, boost::none);

    dynamics(nu, delta, n, delta_c, n_c);
    kinematics(eta, nu);
}
