#include "../include/USV.hpp"
#include <g2o/stuff/misc.h>
#include <iostream>

using namespace teb_local_planner;

USV::USV(double V_current, double beta_current):
    H_rg(6,6), MRB_CG(6,6), MRB(6,6), MA(6,6), M(6,6), Minv(6,6), G_CF(6,6), H(6,6), G(6,6), D(6,6)
{
    g = 9.81;
    rho = 1026;
    
    T_n = 1.0;
    L = 2.0;
    B = 1.08;

    m = 55.0;
    mp = 25.0;
    m_total = m + mp;
    rp << 0, 0, -0.35;
    rg << 0.2, 0, -0.2;
    rg = (m * rg + mp * rp) / (m + mp);
    S_rg = Smtrx(rg);
    H_rg = Hmtrx(rg);
    S_rp = Smtrx(rp);

    R44 = 0.4 * B;
    R55 = 0.25 * L;
    R66 = 0.25 * L;
    T_yaw = 1.0;
    U_max = 6 * 0.5144;

    B_pont = 0.25;
    y_pont = 0.395;
    Cw_pont = 0.75;
    Cb_pont = 0.4;

    nabla = (m + mp) / rho;
    T = nabla / (2 * Cb_pont * B_pont * L);
    Ig_CG = m * (Eigen::Vector3d(pow(R44,2), pow(R55,2), pow(R66,2)).asDiagonal());
    Ig = Ig_CG - m * S_rg * S_rg - mp * S_rp * S_rp;

    l1 = -y_pont;
    l2 = y_pont;
    k_pos = 0.02216 / 2;
    k_neg = 0.01289 / 2;
    n_max = sqrt((0.5 * 24.4 * g) / k_pos);
    n_min = -sqrt((0.5 * 13.6 * g) / k_neg);

    MRB_CG = Eigen::MatrixXd::Zero(6,6);
    MRB_CG.block<3,3>(0,0) = (m + mp) * Eigen::Matrix3d::Identity(3,3);
    MRB_CG.block<3,3>(3,3) = Ig;
    MRB = H_rg.transpose() * MRB_CG * H_rg;

    Xudot = -0.1 * m;
    Yvdot = -1.5 * m;
    Zwdot = -1.0 * m;
    Kpdot = -0.2 * Ig(0,0);
    Mqdot = -0.8 * Ig(1,1);
    Nrdot = -1.7 * Ig(2,2);

    MA = (Eigen::VectorXd(6) << Xudot, Yvdot, Zwdot, Kpdot, Mqdot, Nrdot).finished().asDiagonal();
    MA = -MA;
    M = MRB + MA;
    Minv = M.inverse();

    Aw_pont = Cw_pont * L * B_pont;
    I_T = 2 / 12 * L * pow(B_pont,3) * (6 * pow(Cw_pont,3) / ((1 + Cw_pont) * (1 + 2 * Cw_pont))) + 2 * Aw_pont * pow(y_pont,2);
    I_L = 0.8 * 2 * B_pont * pow(L,3) / 12;
    KB = (5 * T / 2 - 0.5 * nabla / (L * B_pont)) / 3;
    BM_T = I_T / nabla;
    BM_L = I_L / nabla;

    KM_T = KB + BM_T;
    KM_L = KB + BM_L;
    KG = T - rg(2);
    GM_T = KM_T - KG;
    GM_L = KM_L - KG;

    G33 = rho * g * (2 * Aw_pont);
    G44 = rho * g * nabla * GM_T;
    G55 = rho * g * nabla * GM_L;
    G_CF = (Eigen::VectorXd(6) << 0, 0, G33, G44, G55, 0).finished().asDiagonal();
    LCF = -0.2;
    H = Hmtrx(Eigen::Vector3d(LCF, 0.0, 0.0));
    G = H.transpose() * G_CF * H;

    w3 = sqrt(G33 / M(2,2));
    w4 = sqrt(G44 / M(3,3));
    w5 = sqrt(G55 / M(4,4));

    Xu = -24.4 * g / U_max;
    Yv = 0;
    Zw = -2 * 0.3 * w3 * M(2, 2);
    Kp = -2 * 0.2 * w4 * M(3, 3);
    Mq = -2 * 0.4 * w5 * M(4, 4);
    Nr = -M(5, 5) / T_yaw;

    D = (Eigen::VectorXd(6) << Xu, Yv, Zw, Kp, Mq, Nr).finished().asDiagonal();
    D = -D;

    trim_moment = 0;
    trim_setpoint = 0;

    B_config << 1, 1, -l1, -l2;
    B_config = k_pos * B_config;
    Binv = B_config.inverse();

    e_int_psi = 0;
    wn_psi = 4.5;
    zeta_psi = 1;

    e_int_u = 0;
    K_sigma = 15;
    phi = 6;
    
    V_c = V_current;
    beta_c = beta_current;

    sampleTime = 0.02;
}


void USV::dynamics(Eigen::VectorXd& eta, Eigen::VectorXd& nu, Eigen::Vector2d& n, Eigen::Vector2d& n_c)
{
    double u_c = V_c * cos(beta_c - eta(5));
    double v_c = V_c * sin(beta_c - eta(5));

    Eigen::VectorXd  nu_c(6);
    nu_c << u_c, v_c, 0, 0, 0, 0;
    Eigen::VectorXd  nu_r = nu - nu_c;

    Eigen::MatrixXd CRB_CG = Eigen::MatrixXd::Zero(6,6);
    CRB_CG.block<3,3>(0,0) = m_total * Smtrx(nu.tail(3));
    CRB_CG.block<3,3>(3,3) = -Smtrx(Ig * nu.tail(3));
    Eigen::MatrixXd CRB = H_rg.transpose() * CRB_CG * H_rg;

    Eigen::MatrixXd CA = m2c(MA, nu_r);
    CA(5,0) = 0;
    CA(5,1) = 0;

    Eigen::MatrixXd C = CRB + CA;

    Eigen::VectorXd g_0(6);
    g_0 << 0,0,0,0,trim_moment,0;

    // Control forces and moments - with propeller revolution saturation
    Eigen::Vector2d thrust(0,0);
    for (int i = 0; i<2; ++i)
    {
        n(i) = sat(n(i), n_min, n_max);
        if (n(i)>0)
        thrust(i) = k_pos * n(i) * fabs(n(i));
        else
        thrust(i) = k_neg * n(i) * fabs(n(i));
    }

    // Control forces and moments
    Eigen::VectorXd tau(6);
    tau << thrust(0) + thrust(1),
            0, 0, 0, 0,
           -l1 * thrust(0) - l2 * thrust(1);

    // Hydrodynamic linear damping + nonlinear yaw damping
    Eigen::VectorXd tau_damp = - D * nu_r;
    tau_damp(5) += (- 10 * D(5,5) * fabs(nu_r(5)) * nu_r(5));

    // State derivatives (with dimension)
    Eigen::VectorXd tau_crossflow = crossFlowDrag(L, B_pont, T, nu_r);
    Eigen::VectorXd sum_tau = tau + tau_damp + tau_crossflow - C * nu_r - G * eta - g_0;
    Eigen::VectorXd nu_dot = Minv * sum_tau;
    Eigen::Vector2d n_dot = (n_c - n) / T_n;

    nu += sampleTime * nu_dot;
    n += sampleTime * n_dot;
}


double USV::speedControl(double& u, double& u_d, double& u_d_dot)
{
    e_u = u - u_d;
    double sigma = e_u + e_int_u;

    double tau_X = (m_total - Xudot) * (u_d_dot + u_d - u - Xu * u / (m_total-Xudot) - K_sigma * tanh(sigma/phi));

    e_int_u += sampleTime * e_u;

    return tau_X;
}


double USV::headingControl(double& psi, double& psi_d, double& r, double& r_d)
{
    e_psi = g2o::normalize_theta(psi - psi_d);
    double e_r = r - r_d;

    double m = 41.4;
    double T = 1;
    double K = T/m;
    double d = 1/K;
    double k = 0;

    double Kp = m * pow(wn_psi,2) - k;
    double Kd = m * 2 * zeta_psi * wn_psi - d;
    double Ki = (wn_psi / 100) * Kp;

    double tau_N = - Kp * e_psi - Kd * e_r - Ki * e_int_psi;

    e_int_psi += sampleTime * e_psi;

    return tau_N;
}


void USV::controlAllocation(double& tau_X, double& tau_N, Eigen::Vector2d& n_c)
{
    Eigen::Vector2d tau(tau_X, tau_N);
    Eigen::Vector2d u = Binv * tau;

    double n1 = g2o::sign(u(0)) * sqrt(fabs(u(0)));
    double n2 = g2o::sign(u(1)) * sqrt(fabs(u(1)));

    n_c << n1, n2;

    if (n1 > n_max || n1 < n_min || n2 > n_max || n2 < n_min)
        e_int_psi -= sampleTime * e_psi;
        e_int_u -= sampleTime * e_u;

    for (int i = 0; i<2; ++i)
        n_c(i) = sat(n_c(i), n_min, n_max);
}


void USV::kinematics(Eigen::VectorXd& eta, Eigen::VectorXd& nu)
{
    Eigen::Vector3d p_dot = Rzyx(eta(3),eta(4),eta(5)) * nu.head(3);
    Eigen::Vector3d v_dot = Tzyx(eta(3),eta(4)) * nu.tail(3);

    eta.head(3) += sampleTime * p_dot;
    eta.tail(3) += sampleTime * v_dot;
    eta(5) = g2o::normalize_theta(eta(5));
}


void USV::simulate(Eigen::VectorXd& eta, Eigen::VectorXd& nu, Eigen::Vector2d& n, Eigen::Vector2d& n_c, double& psi_d, double& u_d, double& r_d, double& u_d_dot)
{
    double u = nu(0);
    double r = nu(5);
    double psi = eta(5);

    double tau_X = speedControl(u, u_d, u_d_dot);
    double tau_N = headingControl(psi, psi_d, r, r_d);
    
    controlAllocation(tau_X, tau_N, n_c);
    dynamics(eta, nu, n, n_c);
    kinematics(eta, nu);
}